/*
Hook v 0.3 file compressor.
(C) 2007, Nania Francesco Antonio (Italia - Mer (MESSINA))
Modified Ari coder by Nania Francesco Antonio (C) 2007
under GPL, http://www.gnu.org/licenses/gpl.txt
To compile: g++ -O Hook.cpp
This program implements DMC.
Description:
" a state machine in which each state represents a bitwise  context.
Each state has 2 outgoing transitions corresponding to next bits 0 and 1,  
and
a count n0 or n1 associated with each transition. Bit y (0 or 1) is 
compressed
by arithmetic coding with probability ny/(n0+n1) (where ny is n0 or n1 
according
to y), and then ny is incremented. States are cloned (copied) whenever the 
inco-
ming and outgoing counts exceed certain limits. This has the effect of 
creating
a new context extended by 1 bit. In the example  below, the  state  
representing
context 110 is cloned by creating a new state 0110 because the incoming 0 
transi-
tion count (ny for y=0) from state 11 exceeded a limit. This transition is 
moved
to point to the new state. Other incoming transitions (not shown) remain 
pointing
to the original state. The  outgoing transitions are  copied. The  counts of 
the
original state are distributed to the  new state  in  proportion  to  the  
moved
transition's contribution to those counts, which is w = ny/(n0+n1).

                n0 ----> 1100           n0*(1-w) ----> 1100
         ny       /                             /     /
     11 -----> 110                 11        110     /
        (y=0)     \                 |           \   /
                n1 ----> 1101       |   n1*(1-w) ----> 1101
                                    |             /    /
                                    |     n0*w   /    /
                                    | ny        /    /
                                    +-----> 0110    /
                                                \  /
                                          n1*w   --

        Before cloning            After cloning 110 to 0110

In hook v0.1,  the  counts are 32 bit floating point numbers initialized to 
0.1.
The initial state  machine  has 256*255  states  representing  bytewise 
order 1
contexts  with  uniform  statistics.  When  memory  is  exhausted, the model 
is
discarded and the state  machine is  reinitialized. A new state is cloned  
when
ny  >  limit  and  n0+n1-ny  >  length,  where  limit and length are 
parameters.
(Comment by Matt Mahoney :"The Paq data Compression Programs")
*/
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cassert>
#include <stdio.h>
FILE *in;
FILE *out;
unsigned long memsize ;
unsigned char limit,length;
unsigned long x2,x1,xmid,x,bit,filesize,pos,control,bwr,ewr,tot;
unsigned char mbit;
typedef struct emubrain {struct emubrain *ahead[2];float  cxt[2];} branch;
float fct=0.1f;
branch *wizard;
branch *sum;
branch *branchmem;
branch *maxbranch;
branch *addrb;
branch mbuf[1<<16];
/////////////////////// PREDICTOR /////////////////////////////////////////////
int flush()
{

   for (int j=0;j<256;j++)
   for (int i=0;i<255;i++)
   {
   if (i<127)
   {
   mbuf[j+(i<<8)].ahead[0] = &mbuf[j+(((i<<1) + 1)<<8)];
   mbuf[j+(i<<8)].ahead[1] = &mbuf[j+(((i<<1) + 2)<<8)];
   }
   else
   {
   mbuf[j+(i<<8)].ahead[0] = &mbuf[i+1];
   mbuf[j+(i<<8)].ahead[1] = &mbuf[i-127];
   }
   mbuf[j+(i<<8)].cxt[0] = fct;
   mbuf[j+(i<<8)].cxt[1] = fct;
   }

   addrb = branchmem;
   wizard = &mbuf[0];
}
int createbrain(unsigned long memsize)
{
   // creation of DCT structure
   branchmem = (branch*)malloc(memsize);
   if (branchmem == (branch*) NULL)
   {
      fprintf(stderr,"No Memory!\n");
      exit(1);
   }
   maxbranch = branchmem + (memsize/sizeof(branch)) - 20;
   flush();
}

//// PROBABILITY ////////////
double p()
{
   return wizard->cxt[1] / (wizard->cxt[1] +wizard->cxt[0]) ;
}

void update(int y)

{
   float factor;
   if (wizard->cxt[y] >= limit && 
wizard->ahead[y]->cxt[0]+wizard->ahead[y]->cxt[1]>= length + wizard->cxt[y])
   {
   addrb++;
   sum=addrb;
   factor = 
wizard->cxt[y]/(wizard->ahead[y]->cxt[1]+wizard->ahead[y]->cxt[0]);
   sum->cxt[0] = wizard->ahead[y]->cxt[0] * factor;
   sum->cxt[1] = wizard->ahead[y]->cxt[1] * factor;
   wizard->ahead[y]->cxt[0] -= sum->cxt[0];
   wizard->ahead[y]->cxt[1] -= sum->cxt[1];
   sum->ahead[0] = wizard->ahead[y]->ahead[0];
   sum->ahead[1] = wizard->ahead[y]->ahead[1];
   wizard->ahead[y] = sum;
   }
   wizard->cxt[y]++;
   wizard=wizard->ahead[y];
   if (addrb > maxbranch)
   flush();

}
/////////////////////////// ENCODER/DECODER ///////////////////////////////////
int decode()
{
  xmid = x1 + (unsigned long)((x2-x1) * p()); // DOUBLE PRECISION
  if (xmid<x1)xmid=x1;
  if (xmid>=x2) xmid=x2-1;
  int y=0;
  if (x<=xmid) {
    y=1;
    x2=xmid;
  }
  else
    x1=xmid+1;

  update(y);

  // Shift equal MSB's out
  while (((x1^x2)&0xff000000)==0)
  {
    int c;
    x1<<=8;
    x2=(x2<<8)+255;
    c=getc(in);
    x=(x<<8)+c;
  }
return y;
}

void encode(int y)
{
xmid = x1 + (unsigned long)((x2-x1) * p());// DOUBLE PRECISION
  if (xmid<x1) xmid=x1;
  if (xmid>=x2) xmid=x2-1;
  if (y)
    x2=xmid;
  else
    x1=xmid+1;

  update(y);

  // Shift equal MSB's out
  while (((x1^x2)&0xff000000)==0)
  {
    putc(x2>>24, out);
    x1<<=8;
    x2=(x2<<8)+255;

  }

}
void endx()
{
   printf( "Hook v 0.3 file compressor/archiver, (C) 2007 By Nania Francesco Antonio (Italy)\n"
           "This program is free software distributed without warranty under the terms\n"
           "of the GNU General Public License, see http://www.gnu.org/licenses/gpl.txt\n"
           "                              INSTRUCTION                                    \n"
           "To compress:   hook c [memory][limit][lenght] input output \n"
           "To decompress: hook d input output\n"
           "Compression option  Memory needed to compress/decompress\n"
           "------------------  ------------------------------------\n"
           " 0                    8 MB (fastest)\n"
           " 1                   16 MB\n"
           " 2                   32 MB\n"
           " 3                   64 MB\n"
           " 4                  128 MB\n"
           " 5                  256 MB\n"
           " 6                  512 GB\n"
           " 7                    1 GB\n"
           " 8                  1.4 GB\n"
           " 9                  1.8 GB (slowest)\n"
           " Limit =             [1-8]\n"
           " Length option      factor                              \n"
           "------------------  ------------------------------------\n"
           " 0                  1  (for small file)                 \n"
           " 1                  2                                   \n"
           " 2                  3                                   \n"
           " 3                  4                                   \n"
           " 4                  8                                   \n"
           " 5                  16                                  \n"
           " 6                  32                                  \n"
           " 7                  64  (for big file)                  \n"
           " Example compress: Hook c 1 1 7 in out                   \n"
           );
exit(1);
}
//////////////////////////// MAIN ////////////////////////////

int main(int argc, char** argv) {

  // Chech arguments: hook c/d option input output
  if (  argc<4 || argc>7 )
  {
  endx();
  return 1;
  }
  if (  argv[1][0]!='c' && argv[1][0]!='d')
  {
  endx();
  return 1;
  }

  if (argv[1][0]=='c' && (argv[2][0]!='0' && argv[2][0]!='1'&&
  argv[2][0]!='2' && argv[2][0]!='3' && argv[2][0]!='4' && argv[2][0]!='5'&&
  argv[2][0]!='6' && argv[2][0]!='7') && argv[2][0]!='8' && argv[2][0]!='9')
  {
  endx();
  }
   if (argv[1][0]=='c' && (argv[3][0]!='1' && argv[3][0]!='2'&&
  argv[3][0]!='3' && argv[3][0]!='4' && argv[3][0]!='5' && argv[3][0]!='6'&&
  argv[3][0]!='7' && argv[3][0]!='8') )
  {
  endx();
  }
     if (argv[1][0]=='c' && (argv[4][0]!='0' && argv[4][0]!='1'&&
  argv[4][0]!='2' && argv[4][0]!='3' && argv[4][0]!='4' && argv[4][0]!='5'&&
  argv[4][0]!='6' && argv[4][0]!='7') )
  {
  endx();
  }
   if (argv[1][0]=='c' && argc!=7 )
  {
    endx();
  }
    if (argv[1][0]=='d' && argc!=4 )
  {
    endx();
  }

  // Start timer
  clock_t start = clock();

  // Open files
  if (argv[1][0]=='c')
  {
  printf("Hook v 0.3 file compressor/archiver, (C) 2007 \n");
  printf("By Nania Francesco Antonio (Italy)\n");
  if (argv[2][0]=='0') {mbit=23;printf(" Memory = 8 Mb\n");}
  if (argv[2][0]=='1') {mbit=24;printf(" Memory = 16 Mb\n");}
  if (argv[2][0]=='2') {mbit=25;printf(" Memory = 32 Mb\n");}
  if (argv[2][0]=='3') {mbit=26;printf(" Memory = 64 Mb\n");}
  if (argv[2][0]=='4') {mbit=27;printf(" Memory = 128 Mb\n");}
  if (argv[2][0]=='5') {mbit=28;printf(" Memory = 256 Mb\n");}
  if (argv[2][0]=='6') {mbit=29;printf(" Memory = 512 Mb\n");}
  if (argv[2][0]=='7') {mbit=30;printf(" Memory = 1 Gb\n");}
  if (argv[2][0]=='8') {mbit=31;printf(" Memory = 1.4 Gb\n");}
  if (argv[2][0]=='9') {mbit=32;printf(" Memory = 1.8 Gb\n");}
  if (argv[3][0]=='1') {limit=1;}
  if (argv[3][0]=='2') {limit=2;}
  if (argv[3][0]=='3') {limit=3;}
  if (argv[3][0]=='4') {limit=4;}
  if (argv[3][0]=='5') {limit=5;}
  if (argv[3][0]=='6') {limit=6;}
  if (argv[3][0]=='7') {limit=7;}
  if (argv[3][0]=='8') {limit=8;}
  printf(" limit=%d",limit);
  printf(" \n");
  if (argv[4][0]=='0') {length=1;}
  if (argv[4][0]=='1') {length=2;}
  if (argv[4][0]=='2') {length=3;}
  if (argv[4][0]=='3') {length=4;}
  if (argv[4][0]=='4') {length=8;}
  if (argv[4][0]=='5') {length=16;}
  if (argv[4][0]=='6') {length=32;}
  if (argv[4][0]=='7') {length=64;}
  printf(" factor=%d",length);
  printf(" \n");
  in=fopen(argv[5], "rb");
  if (!in) perror(argv[5]), exit(1);
  out=fopen(argv[6], "wb");
  if (!out) perror(argv[6]), exit(1);
  }
  else
  {
  in=fopen(argv[2], "rb");
  if (!in) perror(argv[2]), exit(1);
  out=fopen(argv[3], "wb");
  if (!out) perror(argv[3]), exit(1);
  }

  int c;
  x1=0;
  x2=0xffffffff;
  x=0;
  // Compress
  if (argv[1][0]=='c')
  {
  fseek (in , 0 , SEEK_END);
  filesize = ftell (in);
  rewind (in);
  fwrite (&filesize , 1 , 4 , out);
  fwrite (&mbit , 1 , 1 , out);
  fwrite (&limit , 1 , 1 , out);
  fwrite (&length , 1 , 1 , out);
  if (mbit<31)
  memsize=1<<mbit;
  else
  {
  if (mbit==31) memsize=1400000000;
  if (mbit==32) memsize=1800000000;
  }
  createbrain(memsize);
  while (pos<filesize)
  {
  c=getc(in);
  for (int i=7; i>=0; --i)
  encode((c >> i) &1);
  pos++;
  if (pos>control*(filesize/79))
  {
  printf(">");
  control++;
  }
  }
  for (int i=3; i>=0; --i) //Flush x2
  putc(x2>>(i<<3), out);


  }
  // Decompress
else
{
fread (&filesize , 1 , 4 , in);
fread (&mbit , 1 , 1 , in);
fread (&limit , 1 , 1 , in);
fread (&length , 1 , 1 , in);
if (mbit<31)
memsize=1<<mbit;
else
{
if (mbit==31) memsize=1400000000;
if (mbit==32) memsize=1800000000;
}
createbrain(memsize);
for (int i=3; i>=0; --i) // get x
x+=getc(in)<<(i<<3);

while (pos<filesize)
    {
    c = 0;
    for (int i=7; i>=0; --i)
    c+=c+decode();
    putc(c,out);
    pos++;
    if (pos>control*(filesize/79))
    {
    printf(">");
    control++;
    }
    }
    }

   // Print results
  if (argv[1][0]=='c')
  {
  printf("%s (%ld bytes) compress to %s (%ld bytes) in %1.2f s.\n",
  argv[5], ftell(in), argv[6], ftell(out)
  ,((double)clock()-start)/CLOCKS_PER_SEC);
  }
  else
    {
  printf("%s (%ld bytes) decompress to %s (%ld bytes) in %1.2f s.\n"
  ,argv[2], ftell(in), argv[3], ftell(out)
  ,((double)clock()-start)/CLOCKS_PER_SEC);
  }
  return 0;

}

