/*
Hook v 0.6c Advanced DMC+LZP file compressor.
(C) 2007, Nania Francesco Antonio (Italia - Mer (MESSINA))
->it supports file until 1^64 Bytes;
->include name of file;
->detect extension of file (for LZ , No filter);
Added LZP compression for recursives parts of files (ISO,TXT,DIC ecc.)!
under GPL, http://www.gnu.org/licenses/gpl.txt
To compile: g++ -O Hook.cpp
This program implements DMC.
**************************************************************************
DMC, Copyright 1993, 1987
by Gordon V. Cormack, University of Waterloo, cormack@uwaterloo.ca

All rights reserved.

This code and the algorithms herein are the property of Gordon V. Cormack
and Frank Schwellinger.

Neither the code nor any algorithm herein may be included in any software,
device, or process which is sold, exchanged for profit, or for which a
licence or royalty fee is charged.

Permission is granted to use this code for educational, research, or
commercial purposes, provided this notice is included, and provided this
code is not used as described in the above paragraph.
**************************************************************************
Description:
" a state machine in which each state represents a bitwise  context.
Each state has 2 outgoing transitions corresponding to next bits 0
and 1,  and a count n0 or n1 associated with each transition. Bit y
(0 or 1) is compressed by arithmetic coding with probability ny/(n0+n1)
(where ny is n0 or n1 according to y), and then ny is incremented.
States are cloned (copied) whenever the incoming and outgoing counts
exceed certain limits. This has the effect of creating a new context
extended by 1 bit. In the example  below, the  state  representing
context 110 is cloned by creating a new state 0110 because the
incoming 0 transition count (ny for y=0) from state 11 exceeded a limit.
This transition is moved to point to the new state. Other incoming
transitions (not shown) remain pointing to the original state.
The  outgoing transitions are  copied. The  counts of the original
state are distributed to the  new state  in  proportion  to  the  moved
transition's contribution to those counts, which is w = ny/(n0+n1).

                n0 ----> 1100           n0*(1-w) ----> 1100
         ny       /                             /     /
     11 -----> 110                 11        110     /
        (y=0)     \                 |           \   /
                n1 ----> 1101       |   n1*(1-w) ----> 1101
                                    |             /    /
                                    |     n0*w   /    /
                                    | ny        /    /
                                    +-----> 0110    /
                                                \  /
                                          n1*w   --

        Before cloning            After cloning 110 to 0110

In hook v0.1,  the  counts are 32 bit floating point numbers initialized
to 0.1. The initial state  machine  has 256*255  states  representing
bytewise order 1 contexts  with  uniform  statistics.  When  memory  is
exhausted, the model is discarded and the state  machine is  reinitialized.
A new state is cloned  when ny  >  limit  and  n0+n1-ny  >  length,  where
limit and length are parameters.
(Comment by Matt Mahoney :"The Paq data Compression Programs")
*/
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cassert>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

FILE *in;
FILE *out;
unsigned long long filesize;
unsigned long memsize,maxsize,scanned,umax,rc1,rposo,ubi,rumax,ok,dimu,ramax ;
unsigned long limit=0,avan=0,LZ=0,el1,el2,di1,di2,el3,di3,esis1,esis2;
unsigned long x2,x1,xmid,x,bit,pos,control,tot;
unsigned char mbit; 
char *nf,*extens;;
unsigned long buf[16777216];
unsigned char dct[1<<24][2];
typedef struct emubrain {struct emubrain *ahead[2];float  cxt[2];} branch;
float fct=1.0f;
float pdmc;
branch *wizard;
branch *sum;
branch *branchmem;
branch *maxbranch;
branch *addrb;
branch mbuf[1<<16]; 
               
typedef struct emubrain2 {struct emubrain2 *ahead2[2];float  cxt2[2];} branch2;
branch2 *wizard2;
branch2 *sum2;
branch2 *branchmem2;
branch2 *maxbranch2;
branch2 *addrb2;
branch2 mbuf2[1<<16];
unsigned char *buffer;

unsigned char Peekb(unsigned char *memory)
  {
  unsigned char result;
  memcpy(&result,memory,1);
  return (result);
  }
void Pokeb(unsigned char *memory,unsigned char value)
  {
  memcpy(memory,&value,1);
  }

/////////////////////// PREDICTOR ///////////////////////////////////////
int flush()
{

   for (int j=0;j<256;j++)
   for (int i=0;i<255;i++)
   {
   if (i<127)
   {
   mbuf[j+(i<<8)].ahead[0] = &mbuf[j+(((i<<1) + 1)<<8)];
   mbuf[j+(i<<8)].ahead[1] = &mbuf[j+(((i<<1) + 2)<<8)];
   }
   else
   {
   mbuf[j+(i<<8)].ahead[0] = &mbuf[i+1];
   mbuf[j+(i<<8)].ahead[1] = &mbuf[i-127];
   }
   mbuf[j+(i<<8)].cxt[0] = fct;
   mbuf[j+(i<<8)].cxt[1] = fct;
   }

   addrb = branchmem;
   wizard = &mbuf[0];
}
int createbrain(unsigned long memsize)
{
   // creation of DCT structure
   branchmem = (branch*)malloc(memsize);
   if (branchmem == (branch*) NULL)
   {
      fprintf(stderr,"No Memory!\n");
      exit(1);
   }
   maxbranch = branchmem + (memsize/sizeof(branch)) - 20;
   flush();
}
int flush2()
{

   for (int j=0;j<256;j++)
   for (int i=0;i<255;i++)
   {
   if (i<127)
   {
   mbuf2[j+(i<<8)].ahead2[0] = &mbuf2[j+(((i<<1) + 1)<<8)];
   mbuf2[j+(i<<8)].ahead2[1] = &mbuf2[j+(((i<<1) + 2)<<8)];
   }
   else
   {
   mbuf2[j+(i<<8)].ahead2[0] = &mbuf2[i+1];
   mbuf2[j+(i<<8)].ahead2[1] = &mbuf2[i-127];
   }
   mbuf2[j+(i<<8)].cxt2[0] = fct;
   mbuf2[j+(i<<8)].cxt2[1] = fct;
   }

   addrb2 = branchmem2;
   wizard2 = &mbuf2[0];
}

int createbrain2(unsigned long memsize)
{
   // creation of DCT structure
   branchmem2 = (branch2*)malloc(memsize);
   if (branchmem2 == (branch2*) NULL)
   {
      fprintf(stderr,"No Memory!\n");
      exit(1);
   }
   maxbranch2 = branchmem2 + (memsize/sizeof(branch2)) - 20;
   flush2();
}

//// PROBABILITY ////////////

double bp()
{
return wizard2->cxt2[1] / (wizard2->cxt2[1] +wizard2->cxt2[0]) ;
}

//// PROBABILITY ////////////
double p()
{
return wizard->cxt[1] / (wizard->cxt[1] +wizard->cxt[0]) ;
}

void update(int y)

{
   float factor,length;
   factor=wizard->cxt[y]/(wizard->ahead[y]->cxt[0]+wizard->ahead[y]->cxt[1]);
   length=limit+(wizard->cxt[y]*factor)/(wizard->ahead[y]->cxt[0]+wizard->ahead[y]->cxt[1]);
   // modified update - no length factor ///
   if (wizard->cxt[y] >= length && wizard->ahead[y]->cxt[0]+wizard->ahead[y]->cxt[1]>=(wizard->cxt[y]*factor)*(wizard->cxt[0]+wizard->cxt[1]) ) //***
   {
   addrb++;
   sum=addrb;
   sum->cxt[0] = wizard->ahead[y]->cxt[0] * factor;
   sum->cxt[1] = wizard->ahead[y]->cxt[1] * factor;
   sum->ahead[0] = wizard->ahead[y]->ahead[0];
   sum->ahead[1] = wizard->ahead[y]->ahead[1];
   wizard->ahead[y] = sum;
   }
   wizard->cxt[y]++;
   wizard=wizard->ahead[y];
   if (addrb > maxbranch)
   flush();

}
void bupdate(int y)

{
float factor,length;;
factor=wizard2->cxt2[y]/(wizard2->ahead2[y]->cxt2[0]+wizard2->ahead2[y]->cxt2[1]);
length=limit+(wizard2->cxt2[y]*factor)/(wizard2->ahead2[y]->cxt2[0]+wizard2->ahead2[y]->cxt2[1]);
// modified update - no length factor ///
if (wizard2->cxt2[y] >= length && wizard2->ahead2[y]->cxt2[0]+wizard2->ahead2[y]->cxt2[1]>= (wizard2->cxt2[y]*factor)*(wizard2->cxt2[0]+wizard2->cxt2[1])) //***
{
addrb2++;
sum2=addrb2;
sum2->cxt2[0] = wizard2->ahead2[y]->cxt2[0] * factor;
sum2->cxt2[1] = wizard2->ahead2[y]->cxt2[1] * factor;
sum2->ahead2[0] = wizard2->ahead2[y]->ahead2[0];
sum2->ahead2[1] = wizard2->ahead2[y]->ahead2[1];
wizard2->ahead2[y] = sum2;
}
   wizard2->cxt2[y]++;
   wizard2=wizard2->ahead2[y];
   if (addrb2 > maxbranch2)
   flush2();

}
  double dp() 
  {
    return dct[rc1][1]/(dct[rc1][0]+dct[rc1][1]+0.1);
  }

  void dupdate(int y) 
  {
    if (++dct[rc1][y] > 254) 
    {
     dct[rc1][0] >>= 1;
    dct[rc1][1] >>= 1;
    }

  }
/////////////////////////// ENCODER/DECODER 
///////////////////////////////////
/*
This encoder/decoder is an evolution of the invented arithmetical coder from
Matt Mahoney (c) 2002,
Nania Francesco Antonio (C) 2007.
*/
int decode()
{ 
if (ubi==0) xmid = x1 + (unsigned long)((x2-x1) * p()); // DOUBLE PRECISION
if (ubi==1) xmid = x1 + (unsigned long)((x2-x1) * bp()); // DOUBLE PRECISION
if (ubi==2) xmid = x1 + (unsigned long)((x2-x1) * dp()); // DOUBLE PRECISION
if (xmid<x1)xmid=x1;
if (xmid>=x2) xmid=x2-1;
int y=0;
if (x<=xmid) {y=1;x2=xmid;}
else x1=xmid+1;
if (ubi==0)update(y);
if (ubi==1)bupdate(y);
if (ubi==2)dupdate(y);
// Shift equal MSB's out
while (((x1^x2)&0xff0000)==0)
{
int c;
x1<<=8;
x2=(x2<<8)+255;
c=getc(in);
tot++;
x=(x<<8)+c;
}
return y;
}

void encode(int y)
{
  if (ubi==0) xmid = x1 + (unsigned long)((x2-x1) * p()); // DOUBLE PRECISION
  if (ubi==1) xmid = x1 + (unsigned long)((x2-x1) * bp()); // DOUBLE PRECISION
  if (ubi==2) xmid = x1 + (unsigned long)((x2-x1) * dp()); // DOUBLE PRECISION
  if (xmid<x1) xmid=x1;
  if (xmid>=x2) xmid=x2-1;
  if (y)
  x2=xmid;
  else
  x1=xmid+1;

 if (ubi==0)
  update(y);
  if (ubi==1)
  bupdate(y);
  if (ubi==2)
  dupdate(y);

  // Shift equal MSB's out
  while (((x1^x2)&0xff0000)==0)
  {
    putc(x2>>16, out);
    tot++;
    x1<<=8;
    x2=(x2<<8)+255;

  }

}
void endx()
{
   printf( "Hook v 0.6c file compressor/archiver, (C) 2007 By Nania Francesco A.(Italy)\n"
           "This program is free software distributed without warranty under the terms\n"
           "of the GNU General Public License, see http://www.gnu.org/licenses/gpl.txt\n"
           "                              INSTRUCTION                                 \n"
           "To compress:   hook c [memsize][limit][enable lz]{[lz step]} in out       \n"
           "To decompress: hook d input                                               \n"
           " memsize        -> write MBytes for DMC buffer 1MB=1.000.000 Bytes        \n"
           " Limit          -> write [1-255]                                          \n"
           " LZ compression -> write [0-1] 0=disable  1=enable                        \n"
           " LZ step        -> write [1-255] (if enable LZ!)                          \n"
           " Example   compress: Hook c 256 2 0 in out     (LZ disabled)              \n"
           " Example   compress: Hook c 256 2 1 255 in out (LZ enabled)               \n"
           " Example decompress: Hook d in \n"
           );
exit(1);
}
int Slen( char *f) 
{
  unsigned char result=0;
  while (*(f+result)!=0 )
  result++;
  return result;
}
void writename( char *n) 
{
  int dpoint=0,lex=0;
  extens = (char *)malloc(3);
  unsigned char result=Slen(n);
  unsigned char s;
  fwrite (&result, 1 , 1 , out);
  for (int i=0; i<result; i++)
  {
  s=*(n+i);
  if (dpoint==1) {*(extens+lex)=s;lex++;}
  if (s==0x2e) dpoint=1;
  fwrite (&s, 1 , 1 , out);
  }
  //*(extens+lex)=0;
 // printf("extension=%s \n",extens);
}
void readname() 
{     
  unsigned char result;
  char s;
  int dpoint=0,lex=0;
    extens = (char *)malloc(3);
  fread (&result, 1 , 1 , in);
  nf = (char *)malloc(result);
  for (int i=0; i<result; i++)
  {
  fread (&s, 1 , 1 , in);
  *(nf+i)=s;
  }
  *(nf+result)=0;
  out=fopen(nf, "wb");
  printf("name=%s",nf);
  printf(" \n");
  if (!out) exit(1);
  for (int i=0; i<Slen(nf); i++)
  {
  s=*(nf+i);
  if (dpoint==1) {*(extens+lex)=s;lex++;}
  if (s==0x2e) dpoint=1;
  }
  *(extens+lex)=0;
  
}
void detectf() 
{    
di1=0;di2=0;el2=8;di3=4;el3=16;;esis1=1;esis2=1;// default
if (!strcmp( extens, "log") || !strcmp( extens, "LOG")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("LOG file\n");}
if (!strcmp( extens, "bmp") || !strcmp( extens, "BMP")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("IMAGE BMP\n");}
if (!strcmp( extens, "dic") || !strcmp( extens, "DIC")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=0;printf("Dictionary\n");}
if (!strcmp( extens, "exe") || !strcmp( extens, "EXE")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("EXE file \n");}
if (!strcmp( extens, "hlp") || !strcmp( extens, "HLP")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("Help file \n");}
if (!strcmp( extens, "doc") || !strcmp( extens, "DOC")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("DOC file \n");}
if (!strcmp( extens, "pdf") || !strcmp( extens, "PDF")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("Acrobat Reader \n");}
if (!strcmp( extens, "jpg") || !strcmp( extens, "JPG")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("IMAGE JPEG \n");}
if (!strcmp( extens, "dll") || !strcmp( extens, "DLL")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("EXE Library\n");}
if (!strcmp( extens, "txt") || !strcmp( extens, "TXT")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=0;printf("Text \n");}
if (!strcmp( extens, "iso") || !strcmp( extens, "ISO")) {di1=4;di2=4;el2=4;di3=4;el3=8;esis1=1;esis2=1;printf("ISO image \n");}

}
//////////////////////////// MAIN ////////////////////////////

int main(int argc, char** argv) {

 if (argc >6 && argc<9 && argv[1][0]=='c' && isdigit(*argv[2])) sscanf(argv[2],"%d",&memsize);
 if (argc >6 && argc<9 && argv[1][0]=='c' && isdigit(*argv[3])) sscanf(argv[3],"%d",&limit);
 if (argc >6 && argc<9 && argv[1][0]=='c' && isdigit(*argv[4])) sscanf(argv[4],"%d",&LZ);
 if (argc == 8 && argv[1][0]=='c' && LZ==1 && isdigit(*argv[5])) sscanf(argv[5],"%d",&avan);
  // Chech arguments: hook c/d option input output
  if (  argc<3 || argc>8 ){endx();return 1;}
  if (  argc==7 &&  LZ!=0 && argv[1][0]=='c'){endx();return 1;}
  if (  argc==8 &&  LZ!=1 && argv[1][0]=='c'){endx();return 1;}
  if (  argc!=7 &&  LZ==0 && argv[1][0]=='c'){endx();return 1;}
  if (  argc!=8 &&  LZ==1 && argv[1][0]=='c'){endx();return 1;}
  
  if (  argv[1][0]!='c' && argv[1][0]!='d')
  {
  endx();
  return 1;
  }
  
  if (argv[1][0]=='c' && LZ==1 && (memsize>(1<<31) || limit<1 || avan<1 || limit>255 || avan>255 ))
  {
  endx();
  }
  if (argv[1][0]=='c' && LZ==0 && (memsize>(1<<31) || limit<1  || limit>255   ))
  {
  endx();
  }

   if (argv[1][0]=='c' && LZ==0 && argc!=7 )
  {
    endx();
  }
    if (argv[1][0]=='c' && LZ==1 && argc!=8 )
  {
    endx();
  }
    if (argv[1][0]=='d' && argc!=3 )
  {
    endx();
  }

  // Start timer
  clock_t start = clock();
  printf("Hook v 0.6c Advanced DMC+LZP file compressor/archiver, (C) 2007 \n");
  printf("By Nania Francesco Antonio (Italy)\n");
  // Open files
  if (argv[1][0]=='c' && LZ==1)
  {
  printf(" memory=%d MBytes",memsize," \n");
  printf(" limit=%d",limit," \n");
  printf(" LZ enabled step %d \n",avan);
  in=fopen(argv[6], "rb");
  if (!in) perror(argv[6]), exit(1);
  out=fopen(argv[7], "wb");
  if (!out) perror(argv[7]), exit(1);
  }
  if (argv[1][0]=='c' && LZ==0)
  {
  printf(" memory=%d MBytes",memsize," \n");
  printf(" limit=%d",limit," \n");
  printf(" LZ disabled\n");
  in=fopen(argv[5], "rb");
  if (!in) perror(argv[5]), exit(1);
  out=fopen(argv[6], "wb");
  if (!out) perror(argv[6]), exit(1);
  }
   if (argv[1][0]=='d')
  {
  in=fopen(argv[2], "rb");
  if (!in) perror(argv[2]), exit(1);
  }

  int c;
  x1=0;
  x2=0xffffff;
  x=0;
  maxsize=1<<26;
  buffer = (unsigned char *)malloc(maxsize);

  // Compress
  if (argv[1][0]=='c')
  {
  fseek (in , 0 , SEEK_END);
  filesize = ftell (in);
  rewind (in);
  fwrite (&filesize , 1 , 8 , out);
  fwrite (&memsize , 1 , 4 , out);
  fwrite (&limit , 1 , 4 , out);
  fwrite (&LZ , 1 , 4 , out);
  if (LZ==1) fwrite (&avan , 1 ,4, out);
  if (LZ==1)
  writename(argv[6]);
  else
  writename(argv[5]);
  
  detectf();

  createbrain(memsize*1000000);
  createbrain2(1<<26);
       newscan:
    if (scanned+maxsize>filesize) maxsize=filesize-scanned;
    pos=0;
    umax=0;
    fread (buffer , 1 , maxsize , in);
    
    if (LZ==1)
    {
    memset(buf, 0, sizeof(buf));
    while (pos<maxsize)
    {
     c=(unsigned char)Peekb(buffer+pos);

     if (pos<3 ) // no LZ
      {
      ubi=0;
      for (int i=7; i>=0; --i)
      encode((c>>i)&1); 
      goto l190;
      }

      rc1=(Peekb(buffer+pos-1)>>di1)+(esis1*(Peekb(buffer+pos-2)>>di2)<<el2)+(esis2*(Peekb(buffer+pos-3)>>di3)<<el3);


     if (buf[rc1]>0)
     rposo=buf[rc1]-1;
     else
     {
           ubi=0;
      for (int i=7; i>=0; --i)
      encode((c>>i)&1); 
      goto l180;
        }

     nev:
     if (Peekb(buffer+rposo+umax)==Peekb(buffer+pos+umax) && pos+umax<maxsize)
     {
     umax++;
     goto nev;
     }
     rumax=0;
     dimu=avan;
     ramax=0;
     scal:
     if (rumax+dimu<=umax)
     {
     rumax+=dimu;
     ramax++;
     goto scal;
     }
     umax=rumax;
     l160:
     if (umax==0) // no LZ
      {
      umax=0;
      ubi=2;
      encode(1);
      ubi=0;
      for (int i=7; i>=0; --i)
      encode((c>>i)&1);
      goto l180;
      }
      for (int f=0; f<umax; f++)
      for (int i=7; i>=0; --i)
      update((Peekb(buffer+rposo+f)>>i)&1); 
      ubi=2;
      encode(0); // LZ
      ubi=1;
      umax--;
      if (ramax==1)
      {
      encode(1);
      goto l180;
      }
      rumax=ramax-1;
      nbit:
      encode(0);
      rumax--;
      if (rumax>0) goto nbit;
      encode(1);

        l180:
            if (scanned+pos>control*(filesize/100))
    {
     printf("pack %s %d bytes to %s %d bytes, ratio %f\r",argv[6],(scanned+pos),argv[7],tot,(float)tot/(scanned+pos));
    control++;
    }
        buf[rc1]=pos+1;
        l190:
        pos+=1+umax;
        umax=0;
        }
    }
    else
    {
    while (pos<maxsize)
    {
     c=(unsigned char)Peekb(buffer+pos);
      for (int i=7; i>=0; --i)
      encode((c>>i)&1); // simple o0-o1-o2 // it comes compressed in base to an other analysis of frequency "Udate"
            if (scanned+pos>control*(filesize/100))
    {
     printf("pack %s %d bytes to %s %d bytes, ratio %f\r",argv[5],(scanned+pos),argv[6],tot,(float)tot/(scanned+pos));
    control++;
    }
        pos++;
        }
        }
        
   scanned+=maxsize;

  if (scanned<filesize) goto newscan;

  for (int i=2; i>=0; --i) //Flush x2
  putc(x2>>(i<<3), out);


  }
  // Decompress
else
{
fread (&filesize , 1 , 8 , in);
fread (&memsize , 1 , 4 , in);
fread (&limit , 1 , 4 , in);
fread (&LZ , 1 , 4 , in);
if (LZ==1) fread (&avan , 1 , 4 , in);
readname();
detectf();
createbrain(memsize*1000000);
createbrain2(1<<26);
for (int i=2; i>=0; --i) // get x
x+=getc(in)<<(i<<3);
   
    newdscan:
    if (scanned+maxsize>filesize) maxsize=filesize-scanned;
    pos=0;
    umax=0;
    memset(buf, 0, sizeof(buf));
    if (LZ==1)
     {
    memset(buf, 0, sizeof(buf));
    while (pos<maxsize)
    {
    if (pos<3)
    {
    ubi=0;
    c = 0;
    for (int i=7; i>=0; --i)
    c+=c+decode();
    Pokeb(buffer+pos,c);
    goto nono;
    }
      rc1=(Peekb(buffer+pos-1)>>di1)+(esis1*(Peekb(buffer+pos-2)>>di2)<<el2)+(esis2*(Peekb(buffer+pos-3)>>di3)<<el3);


      if (buf[rc1]==0)
      {
      ubi=0;
       c = 0;
    for (int i=7; i>=0; --i)
    c+=c+decode();
    Pokeb(buffer+pos,c);
      goto l1180;
      }
      ubi=2;
      ok=decode();
      if (ok==1)
      {
      ubi=0;
      c = 0;
    for (int i=7; i>=0; --i)
    c+=c+decode();
    Pokeb(buffer+pos,c);
      goto l1180;
      }
      ubi=1;
      dimu=avan;
      nbyte:
      ok=decode();
      umax+=dimu;
      if (ok==0)
      {
      goto nbyte;
      }

      
      umax--;
      for (int l=0; l<=umax; l++)
      {
      Pokeb(buffer+pos+l,Peekb(buffer+(buf[rc1]-1)+l));
      for (int i=7; i>=0; --i)
      update((Peekb(buffer+(buf[rc1]-1)+l)>>i)&1); 
      }
        l1180:
                          if (scanned+pos>control*(filesize/79))
    {
    printf("unpack %s to %s %d / %d bytes\r",argv[2],nf,scanned+pos,filesize);
    control++;
    }
        buf[rc1]=pos+1;
        nono:
        pos+=1+umax;
        umax=0;
       }
    }
    else
     while (pos<maxsize)
    {

      ubi=0;
       c = 0;
    for (int i=7; i>=0; --i)
    c+=c+decode();
    Pokeb(buffer+pos,c);
    if (scanned+pos>control*(filesize/79))
    {
    printf("unpack %s to %s %d / %d bytes\r",argv[2],nf,scanned+pos,filesize);
    control++;
    }
    pos++;
    }
    fwrite (buffer , 1 , maxsize , out);
    scanned+=maxsize;
    if (scanned<filesize) goto newdscan;
}

   // Print results
  if (argv[1][0]=='c')
  {
  if (LZ==0)
  printf("%s (%ld bytes) compress to %s (%ld bytes) in %1.2f s.\n",
  argv[5], ftell(in), argv[6], ftell(out)
  ,((double)clock()-start)/CLOCKS_PER_SEC);
  else
  printf("%s (%ld bytes) compress to %s (%ld bytes) in %1.2f s.\n",
  argv[6], ftell(in), argv[7], ftell(out)
  ,((double)clock()-start)/CLOCKS_PER_SEC);
  }
  else
    {
  printf("%s (%ld bytes) decompress to %s (%ld bytes) in %1.2f s.\n"
  ,argv[2], ftell(in), nf, ftell(out)
  ,((double)clock()-start)/CLOCKS_PER_SEC);
  }
  return 0;

}
      
